/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.common.binaryfile;

import com.google.protobuf.ByteString;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.common.test.ReflectionTestUtils;
import io.grpc.Context;
import io.grpc.stub.CallStreamObserver;
import lombok.SneakyThrows;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.core.task.TaskExecutor;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.Random;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.function.Function;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

class GrpcBinaryFileServerDownloaderTest {

	@Mock
	private CallStreamObserver<GrpcResponseDummy> callObserver;
	@Mock
	private Function<ByteString, GrpcResponseDummy> chunkBuilder;
	@Mock
	private Consumer<OutputStream> downloadConsumer;
	@Mock
	private TaskExecutor taskExecutor;

	private GrpcBinaryFileServerDownloader<GrpcResponseDummy> downloader;

	@BeforeEach
	void init() {
		downloader = spy(GrpcBinaryFileServerDownloader.<GrpcResponseDummy>builder().callObserver(callObserver).downloadConsumer(downloadConsumer)
				.chunkBuilder(chunkBuilder).taskExecutor(taskExecutor).build());
	}

	@DisplayName("Start")
	@Nested
	class TestStart {

		@BeforeEach
		void init() {
			doNothing().when(downloader).doStart();
		}

		@Test
		void shouldCallDoStart() {

			downloader.start();

			verify(downloader).doStart();
		}

		@Test
		void shouldNotStartWhenStarted() {
			downloader.start();
			downloader.start();

			verify(downloader, times(1)).doStart();
		}
	}

	@Nested
	class TestIsNotStarted {

		@Test
		void shouldReturnTrueIfNotStarted() {
			var result = downloader.isNotStarted();

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalseIfStarted() {
			downloader.isNotStarted();

			var result = downloader.isNotStarted();

			assertThat(result).isFalse();
		}
	}

	@DisplayName("do start")
	@Nested
	class TestDoStart {

		@Captor
		private ArgumentCaptor<Runnable> runnableCaptor;
		@Captor
		private ArgumentCaptor<ExceptionalRunnable> setupStreamCaptor;

		@SneakyThrows
		@Test
		void shouldCallSetupStreams() {
			doNothing().when(downloader).handleSafety(any());

			downloader.doStart();

			verify(downloader).handleSafety(setupStreamCaptor.capture());
			setupStreamCaptor.getValue().run();
			verify(downloader).setupStreams();
		}

		@Test
		void shouldCallStartDownload() {
			doNothing().when(downloader).startDownload();

			downloader.doStart();

			verify(taskExecutor).execute(runnableCaptor.capture());
			runnableCaptor.getValue().run();
			verify(downloader).startDownload();
		}

		@SneakyThrows
		@Test
		void shouldSetOnReadyHandler() {
			downloader.doStart();

			verify(callObserver).setOnReadyHandler(runnableCaptor.capture());
			assertThat(runnableCaptor.getValue()).isNotNull();
		}
	}

	@DisplayName("Start download")
	@Nested
	class TestStartDownload {

		@Captor
		private ArgumentCaptor<ExceptionalRunnable> runnableCaptor;

		@SneakyThrows
		@Test
		void shouldCallDoDownload() {
			doNothing().when(downloader).handleSafety(any());
			doNothing().when(downloader).doDownload();

			downloader.startDownload();

			verify(downloader).handleSafety(runnableCaptor.capture());
			runnableCaptor.getValue().run();
			verify(downloader).doDownload();
		}
	}

	@DisplayName("do")
	@Nested
	class TestDoDownload {

		@Mock
		private PipedOutputStream outputStream;

		@BeforeEach
		void mock() {
			setOutputStreamField(outputStream);
		}

		@SneakyThrows
		@Test
		void shouldCallDownloadConsumer() {
			downloader.doDownload();

			verify(downloadConsumer).accept(outputStream);
		}

		@SneakyThrows
		@Test
		void shouldCloseOutputStream() {
			downloader.doDownload();

			verify(outputStream).close();
		}
	}

	@DisplayName("Send chunks")
	@Nested
	class TestSendChunks {

		@SneakyThrows
		@Test
		void shouldCallHandleSafety() {
			doNothing().when(downloader).doSendChunks();

			downloader.sendChunks();

			verify(downloader).handleSafety(any(ExceptionalRunnable.class));
		}

		@SneakyThrows
		@Test
		void shouldCallDoDownload() {
			doNothing().when(downloader).doSendChunks();

			downloader.sendChunks();

			verify(downloader).doSendChunks();
		}

		@DisplayName("do")
		@Nested
		class TestDoSendChunks {

			@Mock
			private PipedInputStream inputStream;
			@Captor
			private ArgumentCaptor<ByteString> byteStringCaptor;

			private final int readBytes = 20;
			private final byte[] buffer = new byte[readBytes];
			private final GrpcResponseDummy grpcResponseDummy = new GrpcResponseDummy();

			@SneakyThrows
			@BeforeEach
			void mock() {
				doNothing().when(downloader).tryCompleteRequest();
				when(callObserver.isReady()).thenReturn(true);
				when(inputStream.read(any())).thenReturn(readBytes, -1);
				setInputStreamField(inputStream);
				new Random().nextBytes(buffer);
				ReflectionTestUtils.setField(downloader, "buffer", buffer);
			}

			@Test
			void shouldCallChunkBuilder() {
				doSendChunks();

				verify(chunkBuilder).apply(byteStringCaptor.capture());
				assertThat(byteStringCaptor.getValue().toByteArray()).isEqualTo(buffer);
			}

			@DisplayName("should send next chunk if callObserver is ready and stream already received data")
			@Test
			void shouldCallOnNext() {
				when(chunkBuilder.apply(any())).thenReturn(grpcResponseDummy);

				doSendChunks();

				verify(callObserver).onNext(grpcResponseDummy);
			}

			@DisplayName("should call complete grpc stream if download has finished and stream has no data left")
			@Test
			void shouldCallCompleteDownload() {
				setDownloadFinishedField(true);

				doSendChunks();

				verify(downloader).tryCompleteRequest();
			}

			@SneakyThrows
			private void doSendChunks() {
				downloader.doSendChunks();
			}
		}
	}

	@Nested
	class TestTryCompleteRequest {

		@Test
		void shouldCallShouldCompleteRequest() {
			downloader.tryCompleteRequest();

			verify(downloader).shouldCompleteRequest();
		}

		@Test
		void shouldCallCompleteRequest() {
			doReturn(true).when(downloader).shouldCompleteRequest();

			downloader.tryCompleteRequest();

			verify(downloader).completeRequest();
		}

		@Test
		void shouldNotCallCompleteRequest() {
			doReturn(false).when(downloader).shouldCompleteRequest();

			downloader.tryCompleteRequest();

			verify(downloader, never()).completeRequest();
		}
	}

	@Nested
	class TestShouldCompleteRequest {

		@Nested
		class TestWhenDownloadFinished {

			@BeforeEach
			void init() {
				setDownloadFinishedField(true);
			}

			@Test
			void shouldReturnTrue() {
				var result = downloader.shouldCompleteRequest();

				assertThat(result).isTrue();
			}

			@Test
			void shouldReturnFalseIfRequestFinished() {
				setRequestFinishedField(true);

				var result = downloader.shouldCompleteRequest();

				assertThat(result).isFalse();
			}

			@Test
			void shouldUpdateRequestFinished() {
				downloader.shouldCompleteRequest();

				assertThat(getRequestFinished()).isTrue();
			}
		}

		@Nested
		class TestWhenDownloadRunning {

			@BeforeEach
			void init() {
				setDownloadFinishedField(false);
			}

			@Test
			void shouldReturnFalse() {
				var result = downloader.shouldCompleteRequest();

				assertThat(result).isFalse();
			}

			@Test
			void shouldNotUpdateRequestFinished() {
				downloader.shouldCompleteRequest();

				assertThat(getRequestFinished()).isFalse();
			}
		}
	}

	@Nested
	class TestCompleteRequest {

		@Mock
		private PipedInputStream inputStream;

		@BeforeEach
		void mock() {
			setRequestFinishedField(false);
			setDownloadFinishedField(true);
			setInputStreamField(inputStream);
		}

		@SneakyThrows
		@Test
		void shouldCallCloseInputStream() {
			downloader.completeRequest();

			verify(downloader).closeInputStream();
		}

		@Test
		void shouldCallOnCompleted() {
			downloader.completeRequest();

			verify(callObserver).onCompleted();
		}

		@SneakyThrows
		private boolean getRequestFinished() {
			return ReflectionTestUtils.getField(downloader, "requestFinished", AtomicBoolean.class).get();
		}
	}

	@DisplayName("Handle safety")
	@Nested
	class TestHandleSafety {

		@DisplayName("on exception")
		@Nested
		class TestOnException {

			@Mock
			private PipedOutputStream outputStream;
			@Mock
			private PipedInputStream inputStream;

			private final IOException exception = new IOException();

			@SneakyThrows
			@BeforeEach
			void mock() {
				setInputStreamField(inputStream);
				setOutputStreamField(outputStream);
			}

			@SneakyThrows
			@Test
			void shouldThrowTechnicalException() {
				assertThatThrownBy(this::handleSafety).isInstanceOf(TechnicalException.class).extracting(Throwable::getCause).isEqualTo(exception);
			}

			@SneakyThrows
			@Test
			void shouldCloseOutputStream() {
				try {
					handleSafety();
				} catch (Exception e) {
					// do nothing
				}

				verify(outputStream).close();
			}

			@SneakyThrows
			@Test
			void shouldCloseInputStream() {
				try {
					handleSafety();
				} catch (Exception e) {
					// do nothing
				}

				verify(inputStream).close();
			}

			private void handleSafety() {
				downloader.handleSafety(this::dummyMethodThrowingException);
			}

			private void dummyMethodThrowingException() throws IOException {
				throw exception;
			}
		}

	}

	private void setOutputStreamField(OutputStream outputStream) {
		ReflectionTestUtils.setField(downloader, "outputStream", outputStream);
	}

	private void setInputStreamField(InputStream inputStream) {
		ReflectionTestUtils.setField(downloader, "inputStream", inputStream);
	}

	private void setDownloadFinishedField(boolean downloadFinished) {
		ReflectionTestUtils.setField(downloader, "downloadFinished", new AtomicBoolean(downloadFinished));
	}

	private void setRequestFinishedField(boolean requestFinished) {
		ReflectionTestUtils.setField(downloader, "requestFinished", new AtomicBoolean(requestFinished));
	}

	private boolean getRequestFinished() {
		return ReflectionTestUtils.getField(downloader, "requestFinished", AtomicBoolean.class).get();
	}

	static class GrpcResponseDummy {
	}
}